/*
 * Copyright (c) 2010, Stefan Lankes, RWTH Aachen University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *    * Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *    * Neither the name of the University nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @author Stefan Lankes
 * @file include/eduos/tasks.h
 * @brief Task related
 *
 * Create and leave tasks or fork them.
 */

#ifndef __TASKS_H__
#define __TASKS_H__

#include <eduos/stddef.h>
#include <eduos/tasks_types.h>
#include <asm/tasks.h>

#ifdef __cplusplus
extern "C" {
#endif

/** @brief System call to terminate a user level process */
void NORETURN sys_exit(int);

/** @brief Task switcher
 *
 * Timer-interrupted use of this function for task switching
 *
 * @return
 * - 0 no context switch
 * - !0 address of the old stack pointer
 */
size_t** scheduler(void);

/** @brief Initialize the multitasking subsystem
 *
 * This procedure sets the current task to the
 * current "task" (there are no tasks, yet) and that was it.
 *
 * @return
 * - 0 on success
 * - -ENOMEM (-12) on failure
 */
int multitasking_init(void);

/** @brief create a kernel-level task. 
 *
 * @param id The value behind this pointer will be set to the new task's id
 * @param ep Pointer to the entry function for the new task
 * @param args Arguments the task shall start with
 * @param prio Desired priority of the new kernel task
 *
 * @return 
 * - 0 on success
 * - -EINVAL (-22) on failure
 */
int create_kernel_task(tid_t* id, entry_point_t ep, void* args, uint8_t prio);

/** @brief determine the highest priority of all tasks, which are ready
 *
 * @return 
 * - return highest priority
 * - if no task is ready, the function returns an invalid value (> MAX_PRIO)
 */
uint32_t get_highest_priority(void);

/** @brief Call to rescheduling
 *
 * This is a purely assembled procedure for rescheduling
 */
void reschedule(void);

/** @brief Wake up a blocked task
 *
 * The task's status will be changed to TASK_READY 
 *
 * @return
 * - 0 on success
 * - -EINVAL (-22) on failure
 */
int wakeup_task(tid_t);

/** @brief Block current task
 *
 * The current task's status will be changed to TASK_BLOCKED
 *
 * @return
 * - 0 on success
 * - -EINVAL (-22) on failure
 */
int block_current_task(void);

/** @brief Abort current task */
void NORETURN abort(void);

/** @brief This function shall be called by leaving kernel-level tasks */
void NORETURN leave_kernel_task(void);

/** @brief This function shall be called by leaving user-level tasks */
void NORETURN leave_user_task(void);

#ifdef __cplusplus
}
#endif

#endif
